;@echo off
;goto make

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;
; SetKeyboardLeds - Just lights keyboard indicators up
;
;  Written by Four-F (four-f@mail.ru)
;
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

.386
.model flat, stdcall
option casemap:none

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                  I N C L U D E   F I L E S                                        
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

include \masm32\include\w2k\ntstatus.inc
include \masm32\include\w2k\ntddk.inc
include \masm32\include\w2k\ntoskrnl.inc
include \masm32\include\w2k\ntddkbd.inc
include \masm32\include\w2k\hal.inc

includelib \masm32\lib\w2k\ntoskrnl.lib
includelib \masm32\lib\w2k\hal.lib

include \masm32\Macros\Strings.mac

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                       E Q U A T E S                                               
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

WAIT_FOR_ACKNOWLEDGE			equ TRUE
NO_WAIT_FOR_ACKNOWLEDGE			equ FALSE

; Commands to the keyboard (through the 8042 data port).

SET_KEYBOARD_INDICATORS			equ 0EDh
SELECT_SCAN_CODE_SET			equ 0F0h
READ_KEYBOARD_ID				equ 0F2h
SET_KEYBOARD_TYPEMATIC			equ 0F3h
SET_ALL_TYPEMATIC_MAKE_BREAK	equ 0FAh
KEYBOARD_RESET					equ 0FFh

; The default stall value.

I8042_STALL_DEFAULT			equ 50

; The 8042 Controller Status Register bits.

OUTPUT_BUFFER_FULL			equ 01
INPUT_BUFFER_FULL			equ 02
MOUSE_OUTPUT_BUFFER_FULL	equ 20

; Define the i8042 controller input/output ports.

;typedef enum _I8042_IO_PORT_TYPE {
DataPort					equ 0
CommandPort					equ 1
MaximumPortCount			equ 2

; The default allowable retry and polling iterations.

; HKEY_LOCAL_MACHINE\SYSTEM\ControlSet001\Services\i8042prt\Parameters\PollingIterationsMaximum\ResendIterations
I8042_RESEND_DEFAULT		equ 3

; HKEY_LOCAL_MACHINE\SYSTEM\ControlSet001\Services\i8042prt\Parameters\PollingIterations
I8042_POLLING_DEFAULT		equ 12000

; HKEY_LOCAL_MACHINE\SYSTEM\ControlSet001\Services\i8042prt\Parameters\PollingIterationsMaximum
I8042_POLLING_MAXIMUM		equ 12000

; The base port offsets for the i8042 controller command/status and
; data registers.

I8042_DATA_REGISTER		equ 60h
I8042_COMMAND_REGISTER	equ 64h
I8042_STATUS_REGISTER	equ 64h

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                            C O D E                                                
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

.code

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                     PutByteAsynchronous                                           
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

PutByteAsynchronous proc uses esi edi ebx dwPortType:DWORD, dwValue:DWORD

;
; Routine Description:
;     This routine sends a command or data byte to the controller or keyboard
;     asynchronously.  It does not wait for acknowledgment.  If the hardware
;     was not ready for input, the byte is not sent.
;
; Arguments:
;
;     dwPortType - If CommandPort, send the byte to the command register,
;                  otherwise send it to the data register.
;     dwValue    - The byte to send to the hardware. Only LSB is used.
;
; Return Value:
;     None.
;

	; Make sure the Input Buffer Full controller status bit is clear.
	; Time out if necessary.

	xor ebx, ebx
	.while sdword ptr ebx < I8042_POLLING_DEFAULT

		invoke READ_PORT_UCHAR, I8042_STATUS_REGISTER

		and al, INPUT_BUFFER_FULL
		.break .if al != INPUT_BUFFER_FULL
		
		inc ebx				; Try again

	.endw

	.if sdword ptr ebx < I8042_POLLING_DEFAULT

		; Send the byte to the appropriate (command/data) port.

		mov eax, dwValue
		and eax, 0FFh			; Use only LSB
		.if dwPortType == CommandPort
			invoke WRITE_PORT_UCHAR, I8042_COMMAND_REGISTER, eax
		.else
			invoke WRITE_PORT_UCHAR, I8042_DATA_REGISTER, eax
		.endif

    .else
    	invoke DbgPrint, $CTA0("Exceeded number of retries")
    .endif

	ret

PutByteAsynchronous endp

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                  KeyboardAsynchWritePort                                          
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

KeyboardAsynchWritePort proc dwValue:DWORD

;
; Routine Description:
;     This routine writes values to the keyboard device.
;
; Arguments:
;     dwValue - Value to write to the keyboard. Only LSB is used!
;     
; Return Value:
;     None.
;

	invoke PutByteAsynchronous, DataPort, dwValue

	ret

KeyboardAsynchWritePort endp

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                              D I S C A R D A B L E   C O D E                                      
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

.code INIT

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                       DriverEntry                                                 
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

DriverEntry proc pDriverObject:PDRIVER_OBJECT, pusRegistryPath:PUNICODE_STRING

local status:NTSTATUS

	mov status, STATUS_DEVICE_CONFIGURATION_ERROR

	invoke PutByteAsynchronous, DataPort, SET_KEYBOARD_INDICATORS
	invoke KeStallExecutionProcessor, I8042_STALL_DEFAULT
	invoke PutByteAsynchronous, DataPort, 0111y

	mov eax, status
	ret

DriverEntry endp

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
;                                                                                                   
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

end DriverEntry

:make

set drv=SetKeyboardLeds

if exist ..\%drv%.sys del ..\%drv%.sys

\masm32\bin\ml /nologo /c /coff %drv%.bat
\masm32\bin\link /nologo /driver /base:0x10000 /align:32 /out:%drv%.sys /subsystem:native /ignore:4078 %drv%.obj

del %drv%.obj

echo.
pause
